from pathlib import Path
import re

from sqlalchemy.orm import Session
from fastapi.testclient import TestClient

from labelu.internal.common.config import settings
from labelu.internal.domain.models.task import Task
from labelu.internal.adapter.persistence import crud_user
from labelu.internal.adapter.persistence import crud_task
from labelu.internal.adapter.persistence import crud_attachment
from labelu.tests.utils.utils import empty_task_upload


class TestClassTaskAttachmentRouter:
    def test_upload_file_successful(
        self, client: TestClient, testuser_token_headers: dict, db: Session
    ) -> None:

        # prepare data
        data = {
            "name": "task name",
            "description": "task description",
            "tips": "task tips",
        }
        
        task = client.post(
            f"{settings.API_V1_STR}/tasks", headers=testuser_token_headers, json=data
        )
        task_id = task.json()["data"]["id"]
        
        empty_task_upload(task_id, "test.png")

        # run
        with Path("labelu/tests/data/test.png").open(mode="rb") as f:
            png_res = client.post(
                f"{settings.API_V1_STR}/tasks/{task_id}/attachments",
                headers=testuser_token_headers,
                files={"file": ("test.png", f, "image/png")},
            )
        # check
        json = png_res.json()
        assert png_res.status_code == 201
        
        parts = json["data"]["url"].split("/")[-3:]
        assert Path(f"{settings.MEDIA_ROOT}").joinpath("/".join(parts)).exists()
        parts[-1] = "test-thumbnail.png"
        assert Path(f"{settings.MEDIA_ROOT}").joinpath("/".join(parts)).exists()
        
        empty_task_upload(task_id, "test.jsonl")
    
        with Path("labelu/tests/data/test.jsonl").open(mode="rb") as f:
            jsonl_res = client.post(
                f"{settings.API_V1_STR}/tasks/{task_id}/attachments",
                headers=testuser_token_headers,
                files={"file": ("test.jsonl", f, "application/json")},
            )
            
        json = jsonl_res.json()
        assert jsonl_res.status_code == 201
        
        parts = json["data"]["url"].split("/")[-3:]
        assert Path(f"{settings.MEDIA_ROOT}").joinpath("/".join(parts)).exists()
        
    def test_file_exists(
        self, client: TestClient, testuser_token_headers: dict, db: Session
    ) -> None:
         # prepare data
        data = {
            "name": "task name",
            "description": "task description",
            "tips": "task tips",
        }
        task = client.post(
            f"{settings.API_V1_STR}/tasks", headers=testuser_token_headers, json=data
        )
        task_id = task.json()["data"]["id"]

        # upload file
        with Path("labelu/tests/data/test.png").open(mode="rb") as f:
            attachment = client.post(
                f"{settings.API_V1_STR}/tasks/{task_id}/attachments",
                headers=testuser_token_headers,
                files={"file": f},
            )
        
        assert attachment.status_code == 400

    def test_upload_file_when_task_finished(
        self, client: TestClient, testuser_token_headers: dict, db: Session
    ) -> None:

        # prepare data
        task = crud_task.create(
            db=db,
            task=Task(
                name="name",
                description="description",
                tips="tips",
                config="config",
                media_type="IMAGE",
                status="FINISHED",
            ),
        )

        empty_task_upload(task.id, "test.png")
        # run
        with Path("labelu/tests/data/test.png").open(mode="rb") as f:
            new_res = client.post(
                f"{settings.API_V1_STR}/tasks/{task.id}/attachments",
                headers=testuser_token_headers,
                files={"file": f},
            )

        # check
        assert new_res.status_code == 201

    def test_upload_file_not_image(
        self, client: TestClient, testuser_token_headers: dict, db: Session
    ) -> None:

        # prepare data
        data = {
            "name": "task name",
            "description": "task description",
            "tips": "task tips",
        }
        task = client.post(
            f"{settings.API_V1_STR}/tasks", headers=testuser_token_headers, json=data
        )
        task_id = task.json()["data"]["id"]
        
        empty_task_upload(task_id, "test.txt")

        # run
        with Path("labelu/tests/data/test.txt").open(mode="rb") as f:
            new_res = client.post(
                f"{settings.API_V1_STR}/tasks/{task_id}/attachments",
                headers=testuser_token_headers,
                files={"file": ("test.txt", f, "plain/text")},
            )

        # check
        json = new_res.json()
        assert new_res.status_code == 201

        parts = json["data"]["url"].split("/")[-3:]
        assert Path(f"{settings.MEDIA_ROOT}").joinpath("/".join(parts)).exists()
        parts[-1] = parts[-1][:8] + "-test-thumbnail.txt"
        assert not Path(f"{settings.MEDIA_ROOT}").joinpath("/".join(parts)).exists()

    def test_upload_file_when_task_not_found(
        self, client: TestClient, testuser_token_headers: dict, db: Session
    ) -> None:

        # prepare data

        # run
        with Path("labelu/tests/data/test.png").open(mode="rb") as f:
            new_res = client.post(
                f"{settings.API_V1_STR}/tasks/0/attachments",
                headers=testuser_token_headers,
                files={"file": f},
            )

        # check
        assert new_res.status_code == 404
        assert new_res.json()["err_code"] == 50002

    def test_download_file_successful(
        self, client: TestClient, testuser_token_headers: dict, db: Session
    ) -> None:

        # prepare data
        data = {
            "name": "task name",
            "description": "task description",
            "tips": "task tips",
        }
        task = client.post(
            f"{settings.API_V1_STR}/tasks", headers=testuser_token_headers, json=data
        )
        task_id = task.json()["data"]["id"]
        
        empty_task_upload(task_id, "test.png")

        # upload file
        with Path("labelu/tests/data/test.png").open(mode="rb") as f:
            attachment = client.post(
                f"{settings.API_V1_STR}/tasks/{task_id}/attachments",
                headers=testuser_token_headers,
                files={"file": f},
            )
        url = attachment.json()["data"]["url"]

        # run
        r = client.get(
            url,
            headers=testuser_token_headers,
        )

        # check
        assert r.status_code == 200

    def test_download_file_not_found(
        self, client: TestClient, testuser_token_headers: dict, db: Session
    ) -> None:

        # prepare data
        data = {
            "name": "task name",
            "description": "task description",
            "tips": "task tips",
        }
        task = client.post(
            f"{settings.API_V1_STR}/tasks", headers=testuser_token_headers, json=data
        )
        task_id = task.json()["data"]["id"]

        # run
        r = client.get(
            url=f"{settings.SCHEME}://{settings.HOST}:{settings.PORT}{settings.API_V1_STR}/tasks/attachment/upload/1/0",
            headers=testuser_token_headers,
        )

        # check
        assert r.status_code == 404
        assert r.json()["err_code"] == 51001

    def test_task_delete_task_file(
        self, client: TestClient, testuser_token_headers: dict, db: Session
    ) -> None:

        # prepare data
        current_user = crud_user.get_user_by_username(
            db=db, username="test@example.com"
        )
        task = crud_task.create(
            db=db,
            task=Task(
                name="name",
                description="description",
                tips="tips",
                created_by=current_user.id,
                updated_by=current_user.id,
            ),
        )
        
        empty_task_upload(task.id, "test.png")
        
        # upload file
        with Path("labelu/tests/data/test.png").open(mode="rb") as f:
            attachment = client.post(
                f"{settings.API_V1_STR}/tasks/{task.id}/attachments",
                headers=testuser_token_headers,
                files={"file": f},
            )
        attachment_id = attachment.json()["data"]["id"]
        attachment_path = attachment.json()["data"]["url"]

        # run
        data = {"attachment_ids": [attachment_id]}
        r = client.request("delete",
            f"{settings.API_V1_STR}/tasks/{task.id}/attachments",
            headers=testuser_token_headers,
            json=data,
        )

        # check
        deleted = crud_attachment.get(db=db, attachment_id=attachment_id)
        file_full_path = Path(settings.MEDIA_ROOT).joinpath(attachment_path)
        assert r.status_code == 200
        assert not deleted
        assert not file_full_path.exists()

    def test_task_delete_task_file_not_found_task(
        self, client: TestClient, testuser_token_headers: dict, db: Session
    ) -> None:

        # prepare data

        # run
        data = {"attachment_ids": [1]}
        r = client.request("delete",
            f"{settings.API_V1_STR}/tasks/0/attachments",
            headers=testuser_token_headers,
            json=data,
        )

        # check
        assert r.status_code == 404
        assert r.json()["err_code"] == 51001

    def test_task_delete_task_file_not_owner(
        self, client: TestClient, testuser_token_headers: dict, db: Session
    ) -> None:

        # prepare data
        task = crud_task.create(
            db=db,
            task=Task(
                name="name",
                description="description",
                tips="tips",
                created_by=0,
                updated_by=0,
            ),
        )

        # run
        data = {"attachment_ids": [1]}
        r = client.request("delete",
            f"{settings.API_V1_STR}/tasks/{task.id}/attachments",
            headers=testuser_token_headers,
            json=data,
        )

        # check
        assert r.status_code == 403
        assert r.json()["err_code"] == 30001

    def test_partial_content(
        self, client: TestClient, testuser_token_headers: dict, db: Session
    ) -> None:
        # prepare data
        data = {
            "name": "partial content task",
            "description": "task description",
            "tips": "task tips",
        }
        task = client.post(
            f"{settings.API_V1_STR}/tasks", headers=testuser_token_headers, json=data
        )
        task_id = task.json()["data"]["id"]
        
        empty_task_upload(task_id, "test.txt")

        # 上传测试文件
        with Path("labelu/tests/data/test.txt").open(mode="rb") as f:
            attachment = client.post(
                f"{settings.API_V1_STR}/tasks/{task_id}/attachments",
                headers=testuser_token_headers,
                files={"file": ("test.txt", f, "text/plain")},
            )
        url = attachment.json()["data"]["url"]
        
        # 转换为 partial URL
        partial_url = url.replace("/attachment/", "/partial/")
        
        # 测试 Range 请求
        r1 = client.get(
            partial_url,
            headers={**testuser_token_headers, "Range": "bytes=0-9"},
        )
        
        # 检查
        assert r1.status_code == 206  # 部分内容响应码
        assert r1.headers.get("Content-Length") == "10"
        assert r1.headers.get("Content-Range").startswith("bytes 0-9/")
        
        # 测试没有 Range 的请求
        r2 = client.get(
            partial_url,
            headers=testuser_token_headers,
        )
        
        # 检查
        assert r2.status_code == 200
        assert "Accept-Ranges" in r2.headers
        
        # 测试无效的 Range 请求
        r3 = client.get(
            partial_url,
            headers={**testuser_token_headers, "Range": "invalid"},
        )
        
        # 检查 - 应该回退到完整响应
        assert r3.status_code == 200
        
        # 测试超出文件范围的请求
        with Path("labelu/tests/data/test.txt").open(mode="rb") as f:
            file_size = len(f.read())
            
        r4 = client.get(
            partial_url,
            headers={**testuser_token_headers, "Range": f"bytes={file_size + 10}-{file_size + 20}"},
        )
        
        # 检查 - 应该回退到完整响应
        assert r4.status_code == 200

    def test_partial_content_file_not_found(
        self, client: TestClient, testuser_token_headers: dict, db: Session
    ) -> None:
        # 测试不存在的文件
        r = client.get(
            f"{settings.API_V1_STR}/partial/non-existent/file.txt",
            headers=testuser_token_headers,
        )
        
        # 检查
        assert r.status_code == 404
        assert r.json()["err_code"] == 30003

    def test_upload_file_with_hash_symbol(
        self, client: TestClient, testuser_token_headers: dict, db: Session
    ) -> None:
        # prepare data
        data = {
            "name": "hash symbol test",
            "description": "test for files with # in name",
            "tips": "test tips",
        }
        task = client.post(
            f"{settings.API_V1_STR}/tasks", headers=testuser_token_headers, json=data
        )
        task_id = task.json()["data"]["id"]
        
        empty_task_upload(task_id, "test.txt")

        # 创建一个带有井号的临时文件名
        original_filename = "test#file#with#hash.txt"
        
        # 打开测试文件并上传（使用带有井号的文件名）
        with Path("labelu/tests/data/test.txt").open(mode="rb") as f:
            response = client.post(
                f"{settings.API_V1_STR}/tasks/{task_id}/attachments",
                headers=testuser_token_headers,
                files={"file": (original_filename, f, "text/plain")},
            )
        
        # 检查
        json = response.json()
        assert response.status_code == 201
        
        # 验证文件名中的井号被替换为下划线
        assert "#" not in json["data"]["filename"]
        assert "test_file_with_hash.txt" == json["data"]["filename"]
        
        # 验证文件实际存在于文件系统
        parts = json["data"]["url"].split("/")[-3:]
        assert Path(f"{settings.MEDIA_ROOT}").joinpath("/".join(parts)).exists()


class TestAttachmentUtils:
    def test_filename_sanitization(self):
        """测试文件名清理功能，验证井号(#)是否被正确处理"""
        # 模拟 attachment.py 中的文件名清理逻辑
        original_filename = "test#file#with#hash.txt"
        sanitized = re.sub(r'%', '_pct_', original_filename)
        sanitized = re.sub(r'[\\/*?:"<>|#]', '_', sanitized)
        
        # 验证井号被替换为下划线
        assert "#" not in sanitized
        assert "test_file_with_hash.txt" == sanitized
        
        # 测试其他特殊字符也能被正确处理
        complex_filename = r"file\with/special*chars?:and\"<>|#symbols.txt"
        sanitized = re.sub(r'%', '_pct_', complex_filename)
        sanitized = re.sub(r'[\\/*?:"<>|#]', '_', sanitized)
        
        # 检查正确结果（拆分为多个断言便于排除问题）
        assert sanitized.startswith("file_with_special_chars_")
        assert sanitized.endswith("_symbols.txt")
        assert "and" in sanitized
        # 直接检查整个转义序列被正确处理
        assert "\\" not in sanitized



from typing import Union
from datetime import datetime, timedelta

from jose import jwt
from pydantic import BaseModel
from passlib.context import CryptContext
from fastapi.security import HTTPBearer

from labelu.internal.common.config import settings

security = HTTPBearer()

pwd_context = CryptContext(
    schemes=["bcrypt"],
    deprecated="auto",
    bcrypt__truncate_error=False,
)

class AccessToken(BaseModel):
    id: int
    username: str
    exp: Union[datetime, None] = None


def verify_password(plain_password: str, hashed_password: str) -> bool:
    return pwd_context.verify(plain_password, hashed_password)


def get_password_hash(password: str) -> str:
    return pwd_context.hash(password)


# create access token for user login
def create_access_token(
    token: AccessToken, expires_delta: Union[timedelta, None] = None
):
    # update token expire
    if expires_delta:
        expire = datetime.utcnow() + expires_delta
    else:
        expire = datetime.utcnow() + timedelta(minutes=15)
    token.exp = expire

    # generate jwt token
    encoded_jwt = jwt.encode(
        token.dict(),
        settings.PASSWORD_SECRET_KEY,
        algorithm=settings.TOKEN_GENERATE_ALGORITHM,
    )

    return encoded_jwt



from enum import Enum

from loguru import logger
from fastapi.responses import JSONResponse, FileResponse
from fastapi import FastAPI, Request, status, HTTPException
from fastapi.exceptions import RequestValidationError
from sqlalchemy.exc import SQLAlchemyError
from starlette.exceptions import HTTPException as StarletteHTTPException
from labelu.internal.common.config import settings

import pkg_resources
import os

# common init error code
COMMON_INIT_CODE = 30000

# user init error code
USER_INIT_CODE = 40000

# task init error code
TASK_INIT_CODE = 50000

# export
EXPORT_INIT_CODE = 60000

UNEXPECTED_ERROR_CODE = 99999


class ErrorCode(Enum):
    """
    business error code
    """
    UNEXPECTED_ERROR = (
        UNEXPECTED_ERROR_CODE,
        "Internal Error",
    )

    # common init error code
    CODE_30000_SQL_ERROR = (
        COMMON_INIT_CODE,
        "Excute SQL Error",
    )
    CODE_30001_NO_PERMISSION = (
        COMMON_INIT_CODE + 1,
        "Forbidden, No permission",
    )
    CODE_30002_VALIDATION_ERROR = (
        COMMON_INIT_CODE + 2,
        "validation error for request",
    )
    CODE_30003_CLIENT_ERROR = (
        COMMON_INIT_CODE + 3,
        "Error request",
    )

    # user error code
    CODE_40000_USERNAME_OR_PASSWORD_INCORRECT = (
        USER_INIT_CODE,
        "Incorrect username or password",
    )
    CODE_40001_USERNAME_ALREADY_EXISTS = (
        USER_INIT_CODE + 1,
        "Username Already exists in the system",
    )
    CODE_40002_USER_NOT_FOUND = (USER_INIT_CODE + 2, "User not found")
    CODE_40003_CREDENTIAL_ERROR = (USER_INIT_CODE + 3, "Could not validate credentials")
    CODE_40004_NOT_AUTHENTICATED = (USER_INIT_CODE + 4, "Not authenticated")

    # task error code
    CODE_50000_TASK_ERROR = (TASK_INIT_CODE, "Internal Error")
    CODE_50001_TASK_FINISHED_ERROR = (TASK_INIT_CODE + 1, "Task is finished")
    CODE_50002_TASK_NOT_FOUND = (TASK_INIT_CODE + 2, "Task not found")
    CODE_50003_COLLABORATOR_ALREADY_EXISTS = (
        TASK_INIT_CODE + 3,
        "Collaborator already exists")

    CODE_50004_COLLABORATOR_NOT_FOUND = (
        TASK_INIT_CODE + 4,
        "Collaborator not found",
    )
    # task attachment error code
    CODE_51000_CREATE_ATTACHMENT_ERROR = (
        TASK_INIT_CODE + 1000,
        "Upload attachment error, save file failure",
    )
    CODE_51001_TASK_ATTACHMENT_NOT_FOUND = (
        TASK_INIT_CODE + 1001,
        "Attachment file not found",
    )

    CODE_51002_TASK_ATTACHMENT_ALREADY_EXISTS =(
        TASK_INIT_CODE + 1002,
        "Attachment file already exists",
    )
    # task sample error code
    CODE_55000_SAMPLE_LIST_PARAMETERS_ERROR = (
        TASK_INIT_CODE + 5000,
        "Paramenters error: 'after', 'before', 'page' only one must be Ture, page can be 0",
    )
    CODE_55001_SAMPLE_NOT_FOUND = (TASK_INIT_CODE + 5001, "Sample not found")
    CODE_55002_SAMPLE_FORMAT_ERROR = (
        TASK_INIT_CODE + 5002,
        "Sample result format error",
    )
    CODE_55002_SAMPLE_NAME_EXISTS = (
        TASK_INIT_CODE + 5003,
        "Sample name exists",
    )
    CODE_61000_NO_DATA = (
        EXPORT_INIT_CODE + 1000,
        "No data",
    )


class LabelUException(HTTPException):
    def __init__(
        self, code: ErrorCode, status_code=status.HTTP_500_INTERNAL_SERVER_ERROR
    ):
        self.msg = code.value[1]
        self.code = code.value[0]
        self.status_code = status_code


async def labelu_exception_handler(request: Request, exc: LabelUException):
    logger.error(exc)
    return JSONResponse(
        status_code=exc.status_code,
        content={"msg": exc.msg, "err_code": exc.code},
    )


# customize http exception
async def http_exception_handler(request: Request, exc: StarletteHTTPException):
    logger.error(exc)
    if (
        exc.status_code == status.HTTP_404_NOT_FOUND
        and not request.url.path.startswith(settings.API_V1_STR)
    ):
        return FileResponse(
            os.path.join(
                pkg_resources.resource_filename('labelu.internal', 'statics'),
                'index.html'
                ),
            status_code=200,
            headers={'Content-Type': 'text/html', 'Cache-Control': 'no-cache', 'Connection': 'keep-alive'}
            )
    elif exc.status_code == status.HTTP_403_FORBIDDEN:
        JSONResponse(
            status_code=exc.status_code,
            content={
                "msg": str(exc.detail),
                "err_code": ErrorCode.CODE_40004_NOT_AUTHENTICATED.value[0],
            },
        )
    return JSONResponse(
        status_code=exc.status_code,
        content={
            "msg": str(exc.detail),
            "err_code": ErrorCode.CODE_30003_CLIENT_ERROR.value[0],
        },
    )


# only for sqlalchemy
async def sqlalchemy_exception_handler(request: Request, exc: SQLAlchemyError):
    logger.error(exc)
    return JSONResponse(
        status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
        content={
            "msg": str(exc),
            "err_code": ErrorCode.CODE_30000_SQL_ERROR.value[0],
        },
    )


# for http request
async def validation_exception_handler(request, exc):
    logger.error(exc)
    return JSONResponse(
        status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
        content={
            "msg": str(exc),
            "err_code": ErrorCode.CODE_30002_VALIDATION_ERROR.value[0],
        },
    )

async def unexpected_exception_handler(request: Request, exc: Exception):
    logger.error(exc)
    return JSONResponse(
        status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
        content={
            "msg": str(exc),
            "err_code": ErrorCode.UNEXPECTED_ERROR.value[0],
        },
    )

def add_exception_handler(app: FastAPI):
    app.add_exception_handler(RequestValidationError, validation_exception_handler)
    app.add_exception_handler(LabelUException, labelu_exception_handler)
    app.add_exception_handler(StarletteHTTPException, http_exception_handler)
    app.add_exception_handler(SQLAlchemyError, sqlalchemy_exception_handler)
    app.add_exception_handler(Exception, unexpected_exception_handler)



from fastapi import FastAPI

from labelu.internal.middleware.content_type import ContentTypeMiddleware
from labelu.internal.middleware.tracing import TracingMiddleWare


def add_middleware(app: FastAPI):
    app.add_middleware(TracingMiddleWare)
    app.add_middleware(ContentTypeMiddleware)



from typing import List, Union
from pydantic import BaseModel, Field

from labelu.internal.domain.models.task import MediaType


class BasicConfigCommand(BaseModel):
    name: str = Field(description="description: task name", max_length=50)
    media_type: Union[MediaType, None] = Field(
        default=None, description="description: media type of task files"
    )
    description: str = Field(
        default=None, description="description: task description", max_length=500
    )
    tips: Union[str, None] = Field(
        default=None, description="description: task tips", max_length=1000
    )

class CollaboratorUpdateCommand(BaseModel):
    user_id: int = Field(description="description: user id")
    
class CollaboratorBatchCommand(BaseModel):
    user_ids: List[int] = Field(description="description: user ids")

class UpdateCommand(BaseModel):
    name: Union[str, None] = Field(
        default=None, description="description: task name", max_length=50
    )
    description: Union[str, None] = Field(
        default=None, description="description: task description", max_length=500
    )
    tips: Union[str, None] = Field(
        default=None, description="description: task tips", max_length=1000
    )
    media_type: Union[MediaType, None] = Field(
        default=None, description="description: media type of task files"
    )
    config: Union[str, None] = Field(
        default=None, description="description: task config content"
    )



from enum import Enum
import time
from typing import Any, Dict, List
import uuid
from fastapi import WebSocket
from loguru import logger
import asyncio
from dataclasses import dataclass

from pydantic import BaseModel

HEARTBEAT_INTERVAL = 30
HEARTBEAT_TIMEOUT = 60

class MessageType(str, Enum):
    PEERS = "peers"
    PING = "ping"
    PONG = "pong"
    UPDATE = "update"

class Message(BaseModel):
    type: MessageType
    data: Any = None

@dataclass
class ConnectionData:
    id: str = None
    ws: WebSocket = None
    data: Any = None
    heartbeat_task: asyncio.Task = None
    last_heartbeat: float = time.time()
    
    def update_heartbeat(self):
        self.last_heartbeat = time.time()

class ConnectionManager:
    def __init__(self):
        self.active_connections: Dict[str, Dict[str, ConnectionData]] = {}
        
    def _get_connection(self, client_id: str, connection_id: uuid.UUID) -> ConnectionData:
        if client_id in self.active_connections:
            return self.active_connections[client_id].get(connection_id)
        
        return None
            
    async def _heartbeat(self, client_id: str, connection_id: uuid.UUID):
        while True:
            try:
                await asyncio.sleep(HEARTBEAT_INTERVAL)
                
                # Send ping message
                connection = self._get_connection(client_id, connection_id)
                
                if not connection:
                    break
                
                await self.send_message(client_id, connection_id, Message(type=MessageType.PING))
                
                # Check if heartbeat timeout
                if connection and time.time() - connection.last_heartbeat > HEARTBEAT_TIMEOUT:
                    logger.error(f"Heartbeat timeout for client {connection.data.username}")
                    await self.disconnect(client_id, connection.ws)
                    break
                
            except Exception as e:
                logger.error(f"Error in heartbeat for client {client_id}: {e}")
                break
            
    async def connect(self, client_id: str, websocket: WebSocket, data: Any = None) -> ConnectionData:
        await websocket.accept()
        
        if client_id not in self.active_connections:
            self.active_connections[client_id] = []
            
        logger.info(f"Client {client_id} connected")
        id = str(uuid.uuid4())
        connection_data = ConnectionData(
            id=id,
            ws=websocket,
            data=data,
            last_heartbeat=time.time(),
            heartbeat_task=asyncio.create_task(self._heartbeat(client_id, id)),
        )
        
        if not self.active_connections[client_id]:
            self.active_connections[client_id] = {}
            
        self.active_connections[client_id][id] = connection_data
        
        return connection_data
    
    async def disconnect(self, client_id: str, websocket: WebSocket):
        if client_id not in self.active_connections:
            return
        
        connection_to_remove = None
        for connection_id, connection in self.active_connections[client_id].items():
            if connection.ws == websocket:
                connection.heartbeat_task.cancel()
                connection_to_remove = connection
                logger.info(f"Client {client_id} disconnected")
                break
        
        if connection_to_remove is not None:
            del self.active_connections[client_id][connection_to_remove.id]
        
        if client_id in self.active_connections and not self.active_connections[client_id]:
            del self.active_connections[client_id]
        
    async def send_message(self, client_id: str, conn_id: uuid.UUID | None = None, message: Message = None):
        if not conn_id:
            if client_id in self.active_connections:
                for connection in self.active_connections[client_id].values():
                    await connection.ws.send_json(message.dict())
        else:
            connection = self._get_connection(client_id, conn_id)
            if connection:
                await connection.ws.send_json(message.dict())


import xml.etree.ElementTree as ET
from labelu.internal.common.config import settings

class XML_converter:
    def create_pascal_voc_xml(self, config: dict, file: dict, sample_result: dict):
        annotation = ET.Element("annotation")
        
        folder = ET.SubElement(annotation, "folder")
        folder.text = str(settings.MEDIA_ROOT)
        
        filename_elem = ET.SubElement(annotation, "filename")
        filename_elem.text = file.get("filename", "")
        
        path_elem = ET.SubElement(annotation, "path")
        path_elem.text = file.get("path")
        
        image_width = sample_result.get("width", 0)
        image_height = sample_result.get("height", 0)
        
        size_elem = ET.SubElement(annotation, "size")
        width = ET.SubElement(size_elem, "width")
        width.text = str(image_width)
        height = ET.SubElement(size_elem, "height")
        height.text = str(image_height)
        depth = ET.SubElement(size_elem, "depth")
        depth.text = "3"
        rotate = ET.SubElement(size_elem, "rotate")
        rotate.text = str(sample_result.get("rotate", 0))
        
        label_text_dict = {}
        
        for tool in config.get("tools", []):
            label_text_dict[tool.get("tool")] = { attr.get("value"): attr.get("key") for attr in tool.get("config", {}).get("attributes", [])}
        
        common_attributes = { attr.get("value"): attr.get("key") for attr in config.get("attributes", [])}
        
        def get_label(_tool: str, _input_label: str):
            _label = label_text_dict.get(_tool, {}).get(_input_label, "")
            
            if not _label:
                _label = common_attributes.get(_input_label, "")
                
            return _label
        
        for tool in sample_result.copy().keys():
            tool_results = sample_result.pop(tool)
            
            if tool == "rectTool":
                for tool_result in tool_results.get("result", []):
                    
                    obj_elem = ET.SubElement(annotation, "object")
                    name = ET.SubElement(obj_elem, "name")
                    name.text = get_label(tool, tool_result.get("label", ""))
                    # get value from attributes
                    truncated = ET.SubElement(obj_elem, "truncated")
                    truncated_value = tool_result.get("attributes", {}).get("truncated", [0])
                    truncated.text = str(truncated_value[0])
                    difficult = ET.SubElement(obj_elem, "difficult")
                    difficult_value = tool_result.get("attributes", {}).get("difficult", [0])
                    difficult.text = str(difficult_value[0])

                    bndbox = ET.SubElement(obj_elem, "bndbox")
                    xmin = ET.SubElement(bndbox, "xmin")
                    xmin_value = tool_result.get("x", 0)
                    xmin.text = str(xmin_value)
                    ymin = ET.SubElement(bndbox, "ymin")
                    ymin_value = tool_result.get("y", 0)
                    ymin.text = str(ymin_value)
                    xmax = ET.SubElement(bndbox, "xmax")
                    xmax.text = str(xmin_value + tool_result.get("width", 0))
                    ymax = ET.SubElement(bndbox, "ymax")
                    ymax.text = str(ymin_value + tool_result.get("height", 0))
                    
        
        return annotation

    def convert_tool_results(self, tool_name, tool_results):
        annotations = []
        for tool_result in tool_results.get("result", []):
            if tool_name == 'rectTool':
                annotation = self.create_rect(tool_result)
            elif tool_name == 'polygonTool':
                annotation = self.create_polygon(tool_result)
            elif tool_name == 'pointTool':
                annotation = self.create_point(tool_result)
            elif tool_name == 'lineTool':
                annotation = self.create_line(tool_result)
            elif tool_name == 'cuboidTool':
                annotation = self.create_cuboid(tool_result)
            elif tool_name == 'tagTool':
                annotation = self.create_tag(tool_result)
            elif tool_name == 'textTool':
                annotation = self.create_text(tool_result)
            else:
                continue
            annotations.append(annotation)
        return annotations

    def create_attributes(self, attributes: dict):
        result = ET.Element("attributes")
        
        for key, value in attributes.items():
            attribute = ET.SubElement(result, "attribute")
            ET.SubElement(attribute, "key").text = key
            ET.SubElement(attribute, "value").text = value if isinstance(value, str) else ", ".join(value)
        
        return result
    
    def create_rect(self, anno: dict):
        annotation = ET.Element("object")
        
        ET.SubElement(annotation, "toolName").text = "rectTool"
        
        result = ET.Element("result")
        
        ET.SubElement(result, "x").text = str(anno.get("x", 0))
        ET.SubElement(result, "y").text = str(anno.get("y", 0))
        ET.SubElement(result, "width").text = str(anno.get("width", 0))
        ET.SubElement(result, "height").text = str(anno.get("height", 0))
        ET.SubElement(result, "label").text = anno.get("label", "")
        ET.SubElement(result, "order").text = str(anno.get("order", 0))
        
        result.append(self.create_attributes(anno.get("attributes", {})))
        
        annotation.append(result)
        
        return annotation
    
    def create_polygon(self, anno: dict):
        annotation = ET.Element("object")
        
        ET.SubElement(annotation, "toolName").text = "polygonTool"
        
        result = ET.Element("result")
        
        ET.SubElement(result, "label").text = anno.get("label", "")
        ET.SubElement(result, "order").text = str(anno.get("order", 0))
        ET.SubElement(result, "type").text = str(anno.get("type", "line"))
        
        points = ET.Element("points")
        for point in anno.get("points", []):
            p = ET.Element("point")
            
            ET.SubElement(p, "x").text = str(point.get("x", 0))
            ET.SubElement(p, "y").text = str(point.get("y", 0))
            
            points.append(p)
        
        if anno.get("controlPoints"):
            control_points = ET.Element("controlPoints")
            
            for cp in anno.get("controlPoints", []):
                item = ET.Element("controlPoint")
                
                ET.SubElement(item, "x").text = str(cp.get("x", 0))
                ET.SubElement(item, "y").text = str(cp.get("y", 0))
                
                control_points.append(item)
                
            result.append(control_points)
        
        result.append(self.create_attributes(anno.get("attributes", {})))
        result.append(points)
        annotation.append(result)
        
        return annotation
    
    def create_point(self, anno: dict):
        annotation = ET.Element("object")
        
        ET.SubElement(annotation, "toolName").text = "pointTool"
        
        result = ET.Element("result")
        
        ET.SubElement(result, "x").text = str(anno.get("x", 0))
        ET.SubElement(result, "y").text = str(anno.get("y", 0))
        ET.SubElement(result, "label").text = anno.get("label", "")
        ET.SubElement(result, "order").text = str(anno.get("order", 0))
        
        result.append(self.create_attributes(anno.get("attributes", {})))
        annotation.append(result)
        
        return annotation
    
    def create_line(self, anno: dict):
        annotation = ET.Element("object")
        
        ET.SubElement(annotation, "toolName").text = "lineTool"
        
        result = ET.Element("result")
        
        ET.SubElement(result, "label").text = anno.get("label", "")
        ET.SubElement(result, "order").text = str(anno.get("order", 0))
        ET.SubElement(result, "type").text = str(anno.get("type", "line"))
        
        points = ET.Element("points")
        for point in anno.get("points", []):
            p = ET.Element("point")
            
            ET.SubElement(p, "x").text = str(point.get("x", 0))
            ET.SubElement(p, "y").text = str(point.get("y", 0))
            
            points.append(p)
        
        if anno.get("controlPoints"):
            control_points = ET.Element("controlPoints")
            
            for cp in anno.get("controlPoints", []):
                item = ET.Element("controlPoint")
                
                ET.SubElement(item, "x").text = str(cp.get("x", 0))
                ET.SubElement(item, "y").text = str(cp.get("y", 0))
                
                control_points.append(item)
                
            result.append(control_points)
            
        result.append(self.create_attributes(anno.get("attributes", {})))
        result.append(points)
        annotation.append(result)
        
        return annotation
    
    def create_cuboid(self, anno: dict):
        annotation = ET.Element("object")
        
        ET.SubElement(annotation, "toolName").text = "cuboidTool"
        
        result = ET.Element("result")
        
        ET.SubElement(result, "label").text = str(anno.get("label", ""))
        ET.SubElement(result, "order").text = str(anno.get("order", 0))
        ET.SubElement(result, "direction").text = str(anno.get("direction", "front"))
        
        front = ET.Element("front")
        
        for pos, point in anno.get("front", {}).items():
            position = ET.Element(pos)
            ET.SubElement(position, "x").text = str(point.get("x", 0))
            ET.SubElement(position, "y").text = str(point.get("y", 0))
            front.append(position)
        
        back = ET.Element("back")
        for pos, point in anno.get("back", {}).items():
            position = ET.Element(pos)
            ET.SubElement(position, "x").text = str(point.get("x", 0))
            ET.SubElement(position, "y").text = str(point.get("y", 0))
            back.append(position)
            
        result.append(self.create_attributes(anno.get("attributes", {})))
        result.append(front)
        result.append(back)
        
        annotation.append(result)
        
        return annotation
    
    def create_tag(self, anno: dict):
        annotation = ET.Element("object")
        result = ET.SubElement(annotation, "result")
        ET.SubElement(annotation, "toolName").text = "tagTool"
        
        
        ET.SubElement(result, "order").text = str(anno.get("order", 0))
        
        ET.SubElement(result, "id").text = anno.get("id", "")
        
        values = ET.SubElement(result, "values")
        
        for key, value in anno.get("value", {}).items():
            ET.SubElement(values, key).text = ", ".join(value) if value else ""
        
        return annotation
    
    def create_text(self, anno: dict):
        annotation = ET.Element("object")
        
        result = ET.SubElement(annotation, "result")
        ET.SubElement(annotation, "toolName").text = "textTool"
        
        ET.SubElement(result, "order").text = str(anno.get("order", 0))
        ET.SubElement(result, "id").text = anno.get("id", "")
        
        values = ET.SubElement(result, "values")
        
        for key, value in anno.get("value", {}).items():
            ET.SubElement(values, key).text = value
        
        return annotation
  


import re
import aiofiles
import os
from PIL import Image
from pathlib import Path

from loguru import logger
from fastapi import status
from sqlalchemy.orm import Session

from labelu.internal.common.config import settings
from labelu.internal.common.error_code import ErrorCode
from labelu.internal.common.error_code import LabelUException
from labelu.internal.domain.models.user import User
from labelu.internal.domain.models.attachment import TaskAttachment
from labelu.internal.adapter.persistence import crud_task
from labelu.internal.adapter.persistence import crud_attachment
from labelu.internal.application.command.attachment import AttachmentCommand
from labelu.internal.application.command.attachment import AttachmentDeleteCommand
from labelu.internal.application.response.base import CommonDataResp
from labelu.internal.application.response.attachment import AttachmentResponse


async def create(
    db: Session, task_id: int, cmd: AttachmentCommand, current_user: User
) -> AttachmentResponse:

    task = crud_task.get(db=db, task_id=task_id)
    if not task:
        logger.error("cannot find task: {}", task_id)
        raise LabelUException(
            code=ErrorCode.CODE_50002_TASK_NOT_FOUND,
            status_code=status.HTTP_404_NOT_FOUND,
        )

        # file relative path
    path_filename = cmd.file.filename.split("/")
    #  filename = str(uuid.uuid4())[0:8] + "-" + path_filename[-1] NOTE: If you want keep filename safe, you can use uuid as filename 
    filename = path_filename[-1]
    sanitized = re.sub(r'%', '_pct_', filename)
    sanitized = re.sub(r'[\\/*?:"<>|#]', '_', sanitized)
    path = "/".join(path_filename[:-1])
    attachment_relative_base_dir = Path(settings.UPLOAD_DIR).joinpath(
        str(task_id), path
    )
    attachment_relative_path = str(attachment_relative_base_dir.joinpath(sanitized))

    # file full path
    attachment_full_base_dir = Path(settings.MEDIA_ROOT).joinpath(
        attachment_relative_base_dir
    )
    attachment_full_path = Path(settings.MEDIA_ROOT).joinpath(
        attachment_relative_path
    )

    # check file exist
    if attachment_full_path.exists():
        logger.error("file already exists:{}", attachment_full_path)
        raise LabelUException(
            code=ErrorCode.CODE_51002_TASK_ATTACHMENT_ALREADY_EXISTS,
            status_code=status.HTTP_400_BAD_REQUEST,
        )
        
    # create dicreatory
    attachment_full_base_dir.mkdir(parents=True, exist_ok=True)
    
    CHUNK_SIZE = 8 * 1024 * 1024  # 8MB
    logger.info(attachment_full_path)
    try:
        async with aiofiles.open(attachment_full_path, "wb") as out_file:
            total_size = 0
            while True:
                chunk = await cmd.file.read(CHUNK_SIZE)
                if not chunk:
                    break
                await out_file.write(chunk)
                total_size += len(chunk)
                logger.debug(f"{total_size} bytes written")

        logger.info(f"File saved: {attachment_full_path}, size: {total_size} bytes")
    except Exception as e:
        if attachment_full_path.exists():
            os.remove(attachment_full_path)
        logger.error(f"Upload failed: {str(e)}")
        raise LabelUException(
            code=ErrorCode.CODE_51000_CREATE_ATTACHMENT_ERROR,
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            message=f"Upload failed: {str(e)}"
        )

    # create thumbnail for image
    if cmd.file.content_type.startswith("image/"):
        tumbnail_full_path = Path(
            f"{attachment_full_path.parent}/{attachment_full_path.stem}-thumbnail{attachment_full_path.suffix}"
        )
        logger.info(tumbnail_full_path)
        image = Image.open(attachment_full_path)
        image.thumbnail(
            (
                round(image.width / image.height * settings.THUMBNAIL_HEIGH_PIXEL),
                settings.THUMBNAIL_HEIGH_PIXEL,
            ),
        )
        if image.mode != "RGB":
            image = image.convert("RGB")
        image.save(tumbnail_full_path)

    # check file already saved
    if not attachment_full_path.exists() or (
        cmd.file.content_type.startswith("image/") and not tumbnail_full_path.exists()
    ):
        logger.error(
            "cannot find saved images, path is:{}, image content-type is:{}, thumbnail path is:{}",
            attachment_full_path,
            cmd.file.content_type,
            tumbnail_full_path,
        )
        raise LabelUException(
            code=ErrorCode.CODE_51000_CREATE_ATTACHMENT_ERROR,
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
        )

    attachment_url_path = attachment_relative_path.replace("\\", "/")
    attachment_api_url = f"{settings.API_V1_STR}/tasks/attachment/{attachment_url_path}"
    # add a task file record
    with db.begin():
        attachment = crud_attachment.create(
            db=db,
            attachment=TaskAttachment(
                path=attachment_url_path,
                url=attachment_api_url,
                filename=sanitized,
                created_by=current_user.id,
                updated_by=current_user.id,
                task_id=task_id,
            ),
        )

    # response
    return AttachmentResponse(
        id=attachment.id,
        url=attachment_api_url,
        filename=sanitized,
    )


async def download_attachment(file_path: str) -> str:

    # check file exist
    file_full_path = settings.MEDIA_ROOT.joinpath(file_path.lstrip("/"))
    if not file_full_path.is_file() or not file_full_path.exists():
        logger.error("attachment not found:{}", file_full_path)
        raise LabelUException(
            code=ErrorCode.CODE_51001_TASK_ATTACHMENT_NOT_FOUND,
            status_code=status.HTTP_404_NOT_FOUND,
        )

    # response
    return file_full_path


async def delete(
    db: Session, task_id: int, cmd: AttachmentDeleteCommand, current_user: User
) -> CommonDataResp:

    # get task
    task = crud_task.get(db=db, task_id=task_id)
    if not task:
        logger.error("cannot find task:{}", task_id)
        raise LabelUException(
            code=ErrorCode.CODE_51001_TASK_ATTACHMENT_NOT_FOUND,
            status_code=status.HTTP_404_NOT_FOUND,
        )

    if task.created_by != current_user.id:
        logger.error(
            "cannot delete attachment, the task owner is:{}, the delete operator is:{}",
            task.created_by,
            current_user.id,
        )
        raise LabelUException(
            code=ErrorCode.CODE_30001_NO_PERMISSION,
            status_code=status.HTTP_403_FORBIDDEN,
        )

    # delete media
    try:
        attachments = crud_attachment.get_by_ids(
            db=db, attachment_ids=cmd.attachment_ids
        )
        for attachment in attachments:
            file_full_path = Path(settings.MEDIA_ROOT).joinpath(attachment.path)
            os.remove(file_full_path)
    except Exception as e:
        logger.error(e)

    # delete
    with db.begin():
        crud_attachment.delete(db=db, attachment_ids=cmd.attachment_ids)

    # response
    return CommonDataResp(ok=True)



from typing import List
from sqlalchemy.orm import Session
from fastapi import APIRouter, status, Depends, Security
from fastapi.security import HTTPAuthorizationCredentials

from labelu.internal.common import db
from labelu.internal.domain.models.user import User
from labelu.internal.common.security import security
from labelu.internal.dependencies.user import get_current_user
from labelu.internal.application.response.base import MetaData, OkResp, OkRespWithMeta
from labelu.internal.application.service import user as service
from labelu.internal.application.command.user import LoginCommand
from labelu.internal.application.command.user import SignupCommand
from labelu.internal.application.response.user import LoginResponse
from labelu.internal.application.response.user import LogoutResponse
from labelu.internal.application.response.user import UserResponse

router = APIRouter(prefix="/users", tags=["users"])

@router.get('', response_model=OkRespWithMeta[List[UserResponse]], status_code=status.HTTP_200_OK)
async def list_users(
    authorization: HTTPAuthorizationCredentials = Security(security),
    current_user: User = Depends(get_current_user),
    db: Session = Depends(db.get_db),
    page: int | None = 0,
    size: int | None = 10,
    username: str | None = None,
):
    """
    List all users
    """

    # business logic
    data, total = await service.list_by(db=db, page=page, size=size, username=username)
    
    meta_data = MetaData(total=total, page=page, size=len(data))

    # response
    return OkRespWithMeta[List[UserResponse]](meta_data=meta_data, data=data)

@router.get('/me', response_model=OkResp[UserResponse], status_code=status.HTTP_200_OK)
async def get_me(
    authorization: HTTPAuthorizationCredentials = Security(security),
    current_user: User = Depends(get_current_user),
):
    """
    Get current user information
    """

    # response
    return OkResp[UserResponse](data=UserResponse(
        id=current_user.id,
        username=current_user.username,
    ))

@router.post(
    "/signup",
    response_model=OkResp[UserResponse],
    status_code=status.HTTP_201_CREATED,
)
async def signup(cmd: SignupCommand, db: Session = Depends(db.get_db)):
    """
    User signup a account
    """

    # business logic
    data = await service.signup(db=db, cmd=cmd)

    # response
    return OkResp[UserResponse](data=data)


@router.post(
    "/login",
    response_model=OkResp[LoginResponse],
    status_code=status.HTTP_200_OK,
)
async def login(cmd: LoginCommand, db: Session = Depends(db.get_db)):
    """
    User login, get an access token for future requests
    """

    # business logic
    data = await service.login(db=db, cmd=cmd)

    # response
    return OkResp[LoginResponse](data=data)


@router.post(
    "/logout",
    response_model=OkResp[LogoutResponse],
    status_code=status.HTTP_200_OK,
)
async def logout(
    authorization: HTTPAuthorizationCredentials = Security(security),
    db: Session = Depends(db.get_db),
    current_user: User = Depends(get_current_user),
):
    """
    User logout
    """

    # business logic
    pass

    # response
    data = LogoutResponse(msg="succeeded")
    return OkResp[LogoutResponse](data=data)



"""add_collaborators_and_updaters

Revision ID: 2eb983c9a254
Revises: eb9c5b98168b
Create Date: 2025-02-19 16:16:39.259779

"""
from datetime import datetime
from alembic import op
import sqlalchemy as sa
from sqlalchemy.sql import table, column

# revision identifiers, used by Alembic.
revision = '2eb983c9a254'
down_revision = 'eb9c5b98168b'
branch_labels = None
depends_on = None


def upgrade() -> None:
    # Create task_collaborator table
    # if the table is not exists then create it
    conn = op.get_bind()
    inspector = sa.inspect(conn)
    tables = inspector.get_table_names()
    
    if 'task_collaborator' not in tables:
        op.create_table(
            'task_collaborator',
            sa.Column('task_id', sa.Integer(), nullable=False),
            sa.Column('user_id', sa.Integer(), nullable=False),
            sa.Column(
                'created_at',
                sa.DateTime(timezone=True),
                server_default=sa.text('CURRENT_TIMESTAMP'),
                nullable=False
            ),
            sa.ForeignKeyConstraint(
                ['task_id'],
                ['task.id'],
                ondelete='CASCADE'
            ),
            sa.ForeignKeyConstraint(
                ['user_id'],
                ['user.id'],
                ondelete='CASCADE'
            ),
            sa.PrimaryKeyConstraint('task_id', 'user_id')
        )
        
        # Performances index
        indices = inspector.get_indexes('task_collaborator')
        existing_index_names = {idx['name'] for idx in indices}
        
        if 'ix_task_collaborator_task_id' not in existing_index_names:
            op.create_index(
                'ix_task_collaborator_task_id',
                'task_collaborator',
                ['task_id']
            )
            op.create_index(
                'ix_task_collaborator_user_id',
                'task_collaborator',
                ['user_id']
            )
            op.create_index(
                'ix_task_created_by_deleted_at',
                'task',
                ['created_by', 'deleted_at']
            )
    
    # Task sample: updater -> updaters; create a new table task_sample_updater
    if 'task_sample_updater' not in tables:
        op.create_table(
            'task_sample_updater',
            sa.Column('sample_id', sa.Integer(), nullable=False),
            sa.Column('user_id', sa.Integer(), nullable=False),
            sa.Column(
                'created_at',
                sa.DateTime(timezone=True),
                server_default=sa.text('CURRENT_TIMESTAMP'),
                nullable=False
            ),
            sa.ForeignKeyConstraint(
                ['sample_id'],
                ['task_sample.id'],
                ondelete='CASCADE'
            ),
            sa.ForeignKeyConstraint(
                ['user_id'],
                ['user.id'],
                ondelete='CASCADE'
            ),
            sa.PrimaryKeyConstraint('sample_id', 'user_id')
        )
        
        # Performances index
        # check if the index is already exists
        indices = inspector.get_indexes('task_sample_updater')
        existing_index_names = {idx['name'] for idx in indices}
        
        if 'ix_task_sample_updater_sample_id' not in existing_index_names:
            op.create_index(
                'ix_task_sample_updater_sample_id',
                'task_sample_updater',
                ['sample_id']
            )
            op.create_index(
                'ix_task_sample_updater_user_id',
                'task_sample_updater',
                ['user_id']
            )
    
     # Migrate data from task_sample.updated_by to task_sample_updater
    task_sample = table(
        'task_sample',
        column('id', sa.Integer),
        column('updated_by', sa.Integer),
        column('updated_at', sa.DateTime)
    )
    
    task_sample_updater = table(
        'task_sample_updater',
        column('sample_id', sa.Integer),
        column('user_id', sa.Integer),
        column('created_at', sa.DateTime)
    )
    
    conn = op.get_bind()
    for row in conn.execute(sa.select([task_sample.c.id, task_sample.c.updated_by, task_sample.c.updated_at])):
        if row.updated_by:
            conn.execute(
                task_sample_updater.insert().values(
                    sample_id=row.id,
                    user_id=row.updated_by,
                    created_at=row.updated_at or datetime.now()
                )
            )

def downgrade() -> None:
    conn = op.get_bind()
    inspector = sa.inspect(conn)
    tables = inspector.get_table_names()
    
    if 'task_collaborator' in tables:
        indices = inspector.get_indexes('task_collaborator')
        existing_index_names = {idx['name'] for idx in indices}
        
        if 'ix_task_collaborator_user_id' in existing_index_names:
            op.drop_index('ix_task_collaborator_user_id', table_name='task_collaborator')
        if 'ix_task_collaborator_task_id' in existing_index_names:
            op.drop_index('ix_task_collaborator_task_id', table_name='task_collaborator')
            
        op.drop_table('task_collaborator')
    
    if 'task' in tables:
        indices = inspector.get_indexes('task')
        existing_index_names = {idx['name'] for idx in indices}
        
        if 'ix_task_created_by_deleted_at' in existing_index_names:
            op.drop_index('ix_task_created_by_deleted_at', table_name='task')
    
    if 'task_sample_updater' in tables:
        indices = inspector.get_indexes('task_sample_updater')
        existing_index_names = {idx['name'] for idx in indices}
        
        if 'ix_task_sample_updater_user_id' in existing_index_names:
            op.drop_index('ix_task_sample_updater_user_id', table_name='task_sample_updater')
        if 'ix_task_sample_updater_sample_id' in existing_index_names:
            op.drop_index('ix_task_sample_updater_sample_id', table_name='task_sample_updater')
            
        op.drop_table('task_sample_updater')



import json

from alembic import op
import sqlalchemy as sa

def table_exist(table_name):
    """check table is not exist

    Args:
        table_name (string): the name of table

    Returns:
        bool: true or false, whether the table_name exists
    """
    conn = op.get_bind()
    inspector = sa.inspect(conn)
    tables = inspector.get_table_names()
    
    return table_name in tables

def column_exist_in_table(table_name, column_name):
    """check column is not exist in table

    Args:
        table_name (string): the name of table
        column_name (string): the name of column

    Returns:
        bool: true or false, whether the column_name exists in the table_name
    """
    conn = op.get_bind()
    inspector = sa.inspect(conn)
    columns = inspector.get_columns(table_name)
    
    return column_name in [column["name"] for column in columns]


def get_tool_label_dict(task_config: dict) -> dict:
    """get the key value of labels in a given task_id and task_config"""

    label_dict = {"无标签": "noneAttribute"}
    # obtain the labels in current task
    # get the general labels
    for normal_label in task_config.get("attribute", []):
        if normal_label.get("key", ""):
            label_dict[normal_label.get("key")] = normal_label.get("value")
    # get the labels in configuration defined by user
    for task_tool in task_config.get("tools", []):
        if "config" not in task_tool.keys():
            continue
        labels = task_tool.get("config").get("attributeList", [])
        for label in labels:
            if label.get("key", ""):
                label_dict[label.get("key")] = label.get("value")
    return label_dict


def replace_key_with_value(sample_data: dict, label_dict: dict) -> dict:
    """replace the key with value in task_sample table to modify the error for the history version"""

    annotated_result = sample_data.get("result")
    annotated_result = json.loads(annotated_result)
    for sample_tool, sample_tool_results in annotated_result.items():
        if sample_tool.endswith("Tool"):
            for sample_tool_result in sample_tool_results.get("result", []):
                tool_label = sample_tool_result.get("attribute", "")
                if tool_label in label_dict:
                    sample_tool_result["attribute"] = label_dict[tool_label]
    return annotated_result



from datetime import datetime

from typing import List, Union
from pydantic import BaseModel, Field

from labelu.internal.application.response.base import UserResp


class CreatePreAnnotationResponse(BaseModel):
    ids: Union[List[int], None] = Field(
        default=None, description="description: attachment ids"
    )
    
class PreAnnotationFileResponse(BaseModel):
    id: int = Field(
        gt=0,
        description="description: attachment file id",
    )
    filename: str = Field(
        description="description: attachment file name",
    )
    url: str = Field(
        description="description: attachment file url",
    )
    sample_names: List[str] = Field(
        description="description: sample names",
    )


class PreAnnotationResponse(BaseModel):
    id: Union[int, None] = Field(default=None, description="description: annotation id")
    file: Union[object, None] = Field(
        default=None, description="description: media attachment file"
    )
    data: Union[object, None] = Field(
        default=None,
        description="description: sample data, include filename, file url, or result",
    )
    created_at: Union[datetime, None] = Field(
        default=None, description="description: task created at time"
    )
    created_by: Union[UserResp, None] = Field(
        default=None, description="description: task created by"
    )
    updated_at: Union[datetime, None] = Field(
        default=None, description="description: task updated at time"
    )
    updated_by: Union[UserResp, None] = Field(
        default=None, description="description: task updated by"
    )



import shutil
from typing import List, Tuple
from pathlib import Path

from loguru import logger
from fastapi import status
from sqlalchemy.orm import Session

from labelu.internal.common.config import settings
from labelu.internal.common.error_code import ErrorCode
from labelu.internal.common.error_code import LabelUException
from labelu.internal.domain.models.user import User
from labelu.internal.domain.models.task import Task
from labelu.internal.domain.models.task import TaskStatus
from labelu.internal.domain.models.sample import SampleState
from labelu.internal.adapter.persistence import crud_task, crud_user
from labelu.internal.adapter.persistence import crud_sample
from labelu.internal.application.command.task import BasicConfigCommand
from labelu.internal.application.command.task import UpdateCommand
from labelu.internal.application.response.base import UserResp
from labelu.internal.application.response.base import CommonDataResp
from labelu.internal.application.response.task import TaskStatics
from labelu.internal.application.response.task import TaskResponse
from labelu.internal.application.response.task import TaskResponseWithStatics


async def create(
    db: Session, cmd: BasicConfigCommand, current_user: User
) -> TaskResponse:
    # new a task

    with db.begin():
        new_task = crud_task.create(
            db=db,
            task=Task(
                status=TaskStatus.DRAFT.value,
                name=cmd.name,
                description=cmd.description,
                media_type=cmd.media_type,
                tips=cmd.tips,
                created_by=current_user.id,
                updated_by=current_user.id,
            ),
        )

    # response
    return TaskResponse(
        id=new_task.id,
        name=new_task.name,
        description=new_task.description,
        tips=new_task.tips,
        status=new_task.status,
        media_type=new_task.media_type,
        created_at=new_task.created_at,
        created_by=UserResp(
            id=new_task.owner.id,
            username=new_task.owner.username,
        ),
    )


async def list_by(
    db: Session,
    current_user: User,
    page: int,
    size: int,
) -> Tuple[List[TaskResponseWithStatics], int]:
    # get task list
    tasks, total_task = crud_task.list_by(db=db, owner_id=current_user.id, page=page, size=size)

    # get progress
    task_ids = [task.id for task in tasks]
    statics = crud_sample.statics(db=db, task_ids=task_ids)

    # response
    tasks_with_statics = [
        TaskResponseWithStatics(
            id=task.id,
            name=task.name,
            description=task.description,
            tips=task.tips,
            media_type=task.media_type,
            config=task.config,
            status=task.status,
            created_at=task.created_at,
            created_by=UserResp(
                id=task.owner.id,
                username=task.owner.username,
            ),
            stats=TaskStatics(
                new=statics.get(f"{task.id}_{SampleState.NEW.value}", 0),
                done=statics.get(f"{task.id}_{SampleState.DONE.value}", 0),
                skipped=statics.get(f"{task.id}_{SampleState.SKIPPED.value}", 0),
            ),
        )
        for task in tasks
    ]
    return tasks_with_statics, total_task


async def get(db: Session, task_id: int, current_user: User) -> TaskResponseWithStatics:

    # get task detail
    task = crud_task.get(db=db, task_id=task_id)
    
    if not task:
        logger.error("cannot find task:{}", task_id)
        raise LabelUException(
            code=ErrorCode.CODE_50002_TASK_NOT_FOUND,
            status_code=status.HTTP_404_NOT_FOUND,
        )
        
    # not the collaborators
    if task.created_by != current_user.id and current_user not in task.collaborators:
        logger.error(
            "cannot get task, the task owner is:{}, the get operator is:{}",
            task.created_by,
            current_user.id,
        )
        raise LabelUException(
            code=ErrorCode.CODE_30001_NO_PERMISSION,
            status_code=status.HTTP_403_FORBIDDEN,
        )

    # get progress
    statics = crud_sample.statics(
        db=db,
        task_ids=[task.id],
    )

    # response
    return TaskResponseWithStatics(
        id=task.id,
        name=task.name,
        description=task.description,
        tips=task.tips,
        media_type=task.media_type,
        config=task.config,
        status=task.status,
        created_at=task.created_at,
        created_by=UserResp(
            id=task.owner.id,
            username=task.owner.username,
        ),
        stats=TaskStatics(
            new=statics.get(f"{task.id}_{SampleState.NEW.value}", 0),
            done=statics.get(f"{task.id}_{SampleState.DONE.value}", 0),
            skipped=statics.get(f"{task.id}_{SampleState.SKIPPED.value}", 0),
        ),
    )

async def get_collaborators(db: Session, task_id: int, current_user: User):
    task = crud_task.get(db=db, task_id=task_id)
    if not task:
        logger.error("cannot find task:{}", task_id)
        raise LabelUException(
            code=ErrorCode.CODE_50002_TASK_NOT_FOUND,
            status_code=status.HTTP_404_NOT_FOUND,
        )

    return [UserResp(id=user.id, username=user.username) for user in task.collaborators]

async def add_collaborator(db: Session, task_id: int, user_id: int, current_user: User):
    task = crud_task.get(db=db, task_id=task_id)
    if not task:
        logger.error("cannot find task:{}", task_id)
        raise LabelUException(
            code=ErrorCode.CODE_50002_TASK_NOT_FOUND,
            status_code=status.HTTP_404_NOT_FOUND,
        )

    if task.created_by != current_user.id:
        logger.error(
            "cannot add collaborator, the task owner is:{}, the add operator is:{}",
            task.created_by,
            current_user.id,
        )
        raise LabelUException(
            code=ErrorCode.CODE_30001_NO_PERMISSION,
            status_code=status.HTTP_403_FORBIDDEN,
        )

    user = crud_user.get(db=db, id=user_id)
    if not user:
        logger.error("cannot find user:{}", user_id)
        raise LabelUException(
            code=ErrorCode.CODE_40002_USER_NOT_FOUND,
            status_code=status.HTTP_404_NOT_FOUND,
        )
        
    # collaborator exists
    if user in task.collaborators:
        logger.error("collaborator already exists:{}", user_id)
        raise LabelUException(
            code=ErrorCode.CODE_50003_COLLABORATOR_ALREADY_EXISTS,
            status_code=status.HTTP_400_BAD_REQUEST,
        )

    with db.begin():
        task.collaborators.append(user)

    return UserResp(id=user.id, username=user.username)

async def batch_add_collaborators(db: Session, task_id: int, user_ids: List[int], current_user: User):
    task = crud_task.get(db=db, task_id=task_id)
    if not task:
        logger.error("cannot find task:{}", task_id)
        raise LabelUException(
            code=ErrorCode.CODE_50002_TASK_NOT_FOUND,
            status_code=status.HTTP_404_NOT_FOUND,
        )

    if task.created_by != current_user.id:
        logger.error(
            "cannot add collaborator, the task owner is:{}, the add operator is:{}",
            task.created_by,
            current_user.id,
        )
        raise LabelUException(
            code=ErrorCode.CODE_30001_NO_PERMISSION,
            status_code=status.HTTP_403_FORBIDDEN,
        )

    user_ids = set(user_ids) - set([user.id for user in task.collaborators]) - set([task.created_by])
    users = crud_user.list_by_ids(db=db, ids=user_ids)
    if len(users) != len(user_ids):
        logger.error("cannot find users:{}", user_ids)
        raise LabelUException(
            code=ErrorCode.CODE_40002_USER_NOT_FOUND,
            status_code=status.HTTP_404_NOT_FOUND,
        )

    with db.begin():
        task.collaborators.extend(users)

    return [UserResp(id=user.id, username=user.username) for user in users]

async def remove_collaborator(db: Session, task_id: int, user_id: int, current_user: User):
    task = crud_task.get(db=db, task_id=task_id)
    if not task:
        logger.error("cannot find task:{}", task_id)
        raise LabelUException(
            code=ErrorCode.CODE_50002_TASK_NOT_FOUND,
            status_code=status.HTTP_404_NOT_FOUND,
        )

    if task.created_by != current_user.id:
        logger.error(
            "cannot remove collaborator, the task owner is:{}, the remove operator is:{}",
            task.created_by,
            current_user.id,
        )
        raise LabelUException(
            code=ErrorCode.CODE_30001_NO_PERMISSION,
            status_code=status.HTTP_403_FORBIDDEN,
        )

    user = crud_user.get(db=db, id=user_id)
    if not user:
        logger.error("cannot find user:{}", user_id)
        raise LabelUException(
            code=ErrorCode.CODE_40002_USER_NOT_FOUND,
            status_code=status.HTTP_404_NOT_FOUND,
        )

    if user not in task.collaborators:
        logger.error("cannot find collaborator:{}", user_id)
        raise LabelUException(
            code=ErrorCode.CODE_50004_COLLABORATOR_NOT_FOUND,
            status_code=status.HTTP_404_NOT_FOUND,
        )

    with db.begin():
        task.collaborators.remove(user)

    return CommonDataResp(ok=True)

async def batch_remove_collaborators(db: Session, task_id: int, user_ids: List[int], current_user: User):
    task = crud_task.get(db=db, task_id=task_id)
    if not task:
        logger.error("cannot find task:{}", task_id)
        raise LabelUException(
            code=ErrorCode.CODE_50002_TASK_NOT_FOUND,
            status_code=status.HTTP_404_NOT_FOUND,
        )

    if task.created_by != current_user.id:
        logger.error(
            "cannot remove collaborator, the task owner is:{}, the remove operator is:{}",
            task.created_by,
            current_user.id,
        )
        raise LabelUException(
            code=ErrorCode.CODE_30001_NO_PERMISSION,
            status_code=status.HTTP_403_FORBIDDEN,
        )

    users = crud_user.list_by_ids(db=db, ids=user_ids)
    if len(users) != len(user_ids):
        logger.error("cannot find users:{}", user_ids)
        raise LabelUException(
            code=ErrorCode.CODE_40002_USER_NOT_FOUND,
            status_code=status.HTTP_404_NOT_FOUND,
        )

    with db.begin():
        for user in users:
            if user in task.collaborators:
                task.collaborators.remove(user)

    return CommonDataResp(ok=True)

async def update(db: Session, task_id: int, cmd: UpdateCommand) -> TaskResponse:

    # get task
    task = crud_task.get(db=db, task_id=task_id)
    if not task:
        logger.error("cannot find task:{}", task_id)
        raise LabelUException(
            code=ErrorCode.CODE_50002_TASK_NOT_FOUND,
            status_code=status.HTTP_404_NOT_FOUND,
        )

    # update
    obj_in = cmd.dict(exclude_unset=True)
    if cmd.config and cmd.media_type:
        obj_in[Task.status.key] = TaskStatus.CONFIGURED
    else:
        obj_in[Task.config.key] = None
    with db.begin():
        updated_task = crud_task.update(db=db, db_obj=task, obj_in=obj_in)

    # response
    return TaskResponse(
        id=updated_task.id,
        name=updated_task.name,
        description=updated_task.description,
        tips=updated_task.tips,
        media_type=updated_task.media_type,
        config=updated_task.config,
        status=updated_task.status,
        created_at=updated_task.created_at,
        created_by=UserResp(
            id=updated_task.owner.id,
            username=updated_task.owner.username,
        ),
    )


async def delete(db: Session, task_id: int, current_user: User) -> CommonDataResp:

    # get task
    task = crud_task.get(db=db, task_id=task_id)
    if not task:
        logger.error("cannot find task:{}", task_id)
        raise LabelUException(
            code=ErrorCode.CODE_50002_TASK_NOT_FOUND,
            status_code=status.HTTP_404_NOT_FOUND,
        )

    if task.created_by != current_user.id:
        logger.error(
            "cannot delete task, the task owner is:{}, the delete operator is:{}",
            task.created_by,
            current_user.id,
        )
        raise LabelUException(
            code=ErrorCode.CODE_30001_NO_PERMISSION,
            status_code=status.HTTP_403_FORBIDDEN,
        )

    # delete
    with db.begin():
        crud_task.delete(db=db, db_obj=task)

    # delete media
    try:
        file_relative_base_dir = Path(settings.UPLOAD_DIR).joinpath(str(task_id))
        file_full_base_dir = Path(settings.MEDIA_ROOT).joinpath(file_relative_base_dir)
        shutil.rmtree(file_full_base_dir)
    except Exception as e:
        logger.error(e)

    # response
    return CommonDataResp(ok=True)



"""init db

Revision ID: 54fee6a7ecd8
Revises: 
Create Date: 2023-02-28 22:26:57.307010

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '54fee6a7ecd8'
down_revision = None
branch_labels = None
depends_on = None


def upgrade() -> None:
    pass


def downgrade() -> None:
    pass



from datetime import datetime
from typing import Any, Dict, List, Union, Tuple

from sqlalchemy.orm import Session
from sqlalchemy import or_
from fastapi.encoders import jsonable_encoder

from labelu.internal.domain.models.pre_annotation import TaskPreAnnotation


def batch(db: Session, pre_annotations: List[TaskPreAnnotation]) -> List[TaskPreAnnotation]:
    db.bulk_save_objects(pre_annotations, return_defaults=True)
    return pre_annotations


def list_by(
    db: Session,
    task_id: int | None = None,
    sample_name: str | None = None,
    after: int | None = None,
    before: int | None = None,
    page: int | None = None,
    sorting: str | None = None,
    size: int | None = 10,
) -> Tuple[List[TaskPreAnnotation], int]:

    # query filter
    query_filter = [TaskPreAnnotation.deleted_at == None]
    if before:
        query_filter.append(TaskPreAnnotation.id < before)
    if after:
        query_filter.append(TaskPreAnnotation.id > after)
    if task_id:
        query_filter.append(TaskPreAnnotation.task_id == task_id)
        
    if sample_name:
        query_filter.append(or_(TaskPreAnnotation.sample_name == sample_name, TaskPreAnnotation.sample_name == sample_name[9:]))
        
    query = db.query(TaskPreAnnotation).filter(*query_filter)

    # default order by id, before need select last items
    if before:
        query = query.order_by(TaskPreAnnotation.id.desc())
    else:
        query = query.order_by(TaskPreAnnotation.id.asc())
    
    count = query.count()
    
    results = (
        query.offset(offset=page * size if page else 0)
        .limit(limit=size)
        .all()
    )
    
    if sorting:
        field, order = sorting.split(":")
        if order == "desc":
            results = sorted(results, key=lambda x: getattr(x, field), reverse=True)
        else:
            results = sorted(results, key=lambda x: getattr(x, field))
    
    if before:
        results.reverse()
        
    return results, count

def list_by_task_id_and_owner_id(db: Session, task_id: int) -> Dict[str, List[TaskPreAnnotation]]:
    pre_annotations = db.query(TaskPreAnnotation).filter(
        TaskPreAnnotation.task_id == task_id,
        TaskPreAnnotation.deleted_at == None,
    ).all()
    
    return pre_annotations

def list_by_task_id_and_file_id(db: Session, task_id: int, file_id: int) -> List[TaskPreAnnotation]:
    return db.query(TaskPreAnnotation).filter(
        TaskPreAnnotation.task_id == task_id,
        TaskPreAnnotation.deleted_at == None,
        TaskPreAnnotation.file_id == file_id
    ).all()

def list_by_task_id_and_owner_id_and_sample_name(db: Session, task_id: int, sample_name: str) -> List[TaskPreAnnotation]:
    """list pre annotations by task_id, owner_id and sample_name without pagination

    Args:
        db (Session): _description_
        task_id (int): _description_
        owner_id (int): _description_
        sample_name (str): _description_

    Returns:
        List[TaskPreAnnotation]: _description_
    """
    return db.query(TaskPreAnnotation).filter(
        TaskPreAnnotation.task_id == task_id,
        TaskPreAnnotation.deleted_at == None,
        TaskPreAnnotation.sample_name == sample_name
    ).all()

def get(db: Session, pre_annotation_id: int) -> TaskPreAnnotation:
    return (
        db.query(TaskPreAnnotation)
        .filter(TaskPreAnnotation.id == pre_annotation_id, TaskPreAnnotation.deleted_at == None)
        .first()
    )


def get_by_ids(db: Session, pre_annotation_ids: List[int]) -> List[TaskPreAnnotation]:
    return (
        db.query(TaskPreAnnotation)
        .filter(TaskPreAnnotation.id.in_(pre_annotation_ids), TaskPreAnnotation.deleted_at == None)
        .all()
    )


def update(db: Session, db_obj: TaskPreAnnotation, obj_in: Dict[str, Any]) -> TaskPreAnnotation:
    obj_data = jsonable_encoder(obj_in)
    for field in obj_data:
        if field in obj_in:
            setattr(db_obj, field, obj_in[field])
    db.add(db_obj)
    db.flush()
    db.refresh(db_obj)
    return db_obj


def delete(db: Session, pre_annotation_ids: List[int]) -> None:
    db.query(TaskPreAnnotation).filter(TaskPreAnnotation.id.in_(pre_annotation_ids)).update(
        {TaskPreAnnotation.deleted_at: datetime.now()}
    )


def count(db: Session, task_id: int, sample_name: str | None) -> int:
    query_filter = [TaskPreAnnotation.deleted_at == None]
    if task_id:
        query_filter.append(TaskPreAnnotation.task_id == task_id)
        
    if sample_name:
        query_filter.append(TaskPreAnnotation.sample_name == sample_name)
        
    return db.query(TaskPreAnnotation).filter(*query_filter).count()



"""remove_task_attachment_ids

Revision ID: 034c7045b540
Revises: 2eb983c9a254
Create Date: 2025-03-13 22:43:30.564428

"""
from alembic import op

from labelu.alembic_labelu.alembic_labelu_tools import column_exist_in_table

# revision identifiers, used by Alembic.
revision = '034c7045b540'
down_revision = '2eb983c9a254'
branch_labels = None
depends_on = None


def upgrade() -> None:
    # remove task_attachment_ids column from task sample table
    # fix the unnecessary column from: https://github.com/opendatalab/labelU/blob/v1.2.1/labelu/alembic_labelu/versions/2eb983c9a254_add_collaborators_and_updaters.py#L143
    if column_exist_in_table('task_sample', 'task_attachment_ids'):
        op.drop_column('task_sample', 'task_attachment_ids')


def downgrade() -> None:
    if not column_exist_in_table('task_sample', 'task_attachment_ids'):
        op.execute('ALTER TABLE task_sample ADD COLUMN task_attachment_ids TEXT COMMENT "task attachment ids"')



from pathlib import Path
import re
import aiofiles
from sqlalchemy.orm import Session
from fastapi import APIRouter, status, Depends, Security
from fastapi import File, Header, UploadFile
from fastapi.responses import FileResponse, StreamingResponse
from fastapi.security import HTTPAuthorizationCredentials
import mimetypes

from labelu.internal.common import db
from labelu.internal.common.error_code import ErrorCode, LabelUException
from labelu.internal.common.security import security
from labelu.internal.domain.models.user import User
from labelu.internal.dependencies.user import get_current_user
from labelu.internal.application.service import attachment as service
from labelu.internal.application.command.attachment import AttachmentCommand
from labelu.internal.application.command.attachment import AttachmentDeleteCommand
from labelu.internal.application.response.base import OkResp
from labelu.internal.application.response.base import CommonDataResp
from labelu.internal.application.response.attachment import AttachmentResponse


router = APIRouter(prefix="/tasks", tags=["attachments"])


@router.post(
    "/{task_id}/attachments",
    response_model=OkResp[AttachmentResponse],
    status_code=status.HTTP_201_CREATED,
)
async def create(
    task_id: int,
    file: UploadFile = File(...),
    authorization: HTTPAuthorizationCredentials = Security(security),
    db: Session = Depends(db.get_db),
    current_user: User = Depends(get_current_user),
):
    """
    Create attechment as annnotation sample.
    """
    # business logic
    cmd = AttachmentCommand(file=file)
    data = await service.create(
        db=db, task_id=task_id, cmd=cmd, current_user=current_user
    )

    # response
    return OkResp[AttachmentResponse](data=data)

@router.get(
    "/attachment/{file_path:path}",
    response_class=FileResponse,
    status_code=status.HTTP_200_OK,
)
async def download_attachment(file_path: str):
    """
    download attachment.
    """

    # business logic
    data = await service.download_attachment(file_path=file_path)
    
    return data

@router.get(
    "/partial/{file_path:path}",
    response_class=FileResponse,
    status_code=status.HTTP_200_OK,
)
async def get_content(file_path: str, range: str = Header(None)):
    """
    partial content
    """
    
    try:
        full_path = await service.download_attachment(file_path=file_path)
        full_path = Path(full_path) 
    except Exception:
        raise LabelUException(
            code=ErrorCode.CODE_51001_TASK_ATTACHMENT_NOT_FOUND,
            status_code=status.HTTP_404_NOT_FOUND,
        )
    
    if not full_path.exists():
        raise LabelUException(
            code=ErrorCode.CODE_51001_TASK_ATTACHMENT_NOT_FOUND,
            status_code=status.HTTP_404_NOT_FOUND,
        )

    # Business logic
    file_size = full_path.stat().st_size
    media_type = mimetypes.guess_type(str(full_path))[0] or "application/octet-stream"
    
    if not range:
        return FileResponse(
            path=str(full_path),
            media_type=media_type,
            headers={"Accept-Ranges": "bytes"}
        )
        
    try:
        range_match = re.match(r'bytes=(\d+)-(\d*)', range)
        if not range_match:
            raise ValueError("Invalid range format")
        
        start = int(range_match.group(1))
        end_str = range_match.group(2)
        end = int(end_str) if end_str else min(start + 1024 * 1024, file_size - 1)
        
        # 验证范围
        if start < 0 or start >= file_size:
            raise ValueError(f"Start position {start} out of bounds")
        
        if end >= file_size:
            end = file_size - 1
        
        content_length = end - start + 1
        
    except ValueError:
        return FileResponse(
            path=str(full_path),
            media_type=media_type,
            headers={"Accept-Ranges": "bytes"}
        )
        
        
    async def file_stream():
        async with aiofiles.open(str(full_path), 'rb') as f:
            await f.seek(start)
            remaining = content_length
            chunk_size = min(64 * 1024, remaining)
            
            while remaining > 0:
                chunk = await f.read(min(chunk_size, remaining))
                if not chunk:
                    break
                remaining -= len(chunk)
                yield chunk
    
    headers = {
        "Accept-Ranges": "bytes",
        "Content-Range": f"bytes {start}-{end}/{file_size}",
        "Content-Length": str(content_length),
    }
    
    return StreamingResponse(
        file_stream(),
        status_code=206,
        media_type=media_type,
        headers=headers
    )


@router.delete(
    "/{task_id}/attachments",
    response_model=OkResp[CommonDataResp],
    status_code=status.HTTP_200_OK,
)
async def delete(
    task_id: int,
    cmd: AttachmentDeleteCommand,
    authorization: HTTPAuthorizationCredentials = Security(security),
    db: Session = Depends(db.get_db),
    current_user: User = Depends(get_current_user),
):
    """
    delete task.
    """

    # business logic
    data = await service.delete(
        db=db, task_id=task_id, cmd=cmd, current_user=current_user
    )

    # response
    return OkResp[CommonDataResp](data=data)



from fastapi import FastAPI

from labelu.internal.common.config import settings
from labelu.internal.adapter.routers import user
from labelu.internal.adapter.routers import task
from labelu.internal.adapter.routers import sample
from labelu.internal.adapter.routers import attachment
from labelu.internal.adapter.routers import pre_annotation


def add_router(app: FastAPI):
    app.include_router(user.router, prefix=settings.API_V1_STR)
    app.include_router(task.router, prefix=settings.API_V1_STR)
    app.include_router(attachment.router, prefix=settings.API_V1_STR)
    app.include_router(sample.router, prefix=settings.API_V1_STR)
    app.include_router(pre_annotation.router, prefix=settings.API_V1_STR)



import pytest
from typing import Dict, Generator

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from fastapi.testclient import TestClient

from labelu.main import app
from labelu.internal.common.config import settings
from labelu.internal.common.db import Base
from labelu.internal.common.db import get_db
from labelu.internal.common.security import get_password_hash
from labelu.internal.domain.models.user import User
from labelu.internal.adapter.persistence import crud_user

TEST_USERNAME = "test@example.com"
TEST_USER_PASSWORD = "test@123"


SQLALCHEMY_DATABASE_URL = "sqlite:///./test.db"
engine = create_engine(
    SQLALCHEMY_DATABASE_URL, connect_args={"check_same_thread": False}, echo=True
)
TestingSessionLocal = sessionmaker(autocommit=True, autoflush=False, bind=engine)
Base.metadata.create_all(bind=engine)


def override_get_db():
    try:
        db = TestingSessionLocal()
        yield db
    finally:
        db.close()


app.dependency_overrides[get_db] = override_get_db


@pytest.fixture(autouse=True, scope="session")
def db() -> Generator:
    try:
        db = TestingSessionLocal()
        yield db
    finally:
        db.close()


@pytest.fixture(scope="session", autouse=True)
def ensure_test_user() -> Generator:
    # Initialize test user once per test session
    init_db()
    yield


@pytest.fixture(autouse=True)
def run_around_tests():
    # Code that will run before your test, for example:
    print("start up")
    # A test function will be run at this point
    yield
    # Code that will run after your test, for example:
    print("tear down")
    try:
        db = TestingSessionLocal()
        meta = Base.metadata
        with db.begin():
            for table in reversed(meta.sorted_tables):
                if table.key != "user":
                    db.execute(table.delete())
    finally:
        db.close()


@pytest.fixture(scope="module")
def client() -> Generator:
    with TestClient(app) as c:
        yield c


@pytest.fixture(scope="module")
def testuser_token_headers(client: TestClient) -> Dict[str, str]:
    return get_testuser_token_headers(client)


def get_testuser_token_headers(client: TestClient) -> Dict[str, str]:
    data = {
        "username": TEST_USERNAME,
        "password": TEST_USER_PASSWORD,
    }
    r = client.post(f"{settings.API_V1_STR}/users/login", json=data)
    response_json = r.json()
    
    # Add debugging info for CI failures
    if "data" not in response_json:
        print(f"Login failed. Response: {response_json}")
        print(f"Status code: {r.status_code}")
        print(f"Password length (bytes): {len(TEST_USER_PASSWORD.encode('utf-8'))}")
        raise ValueError(f"Login failed with response: {response_json}")
    
    token = response_json["data"]["token"]
    headers = {"Authorization": f"{token}"}
    return headers


def init_db() -> None:
    try:
        db = TestingSessionLocal()
        with db.begin():
            user = crud_user.get_user_by_username(db, username=TEST_USERNAME)
            if not user:
                user_in = User(
                    username=TEST_USERNAME,
                    hashed_password=get_password_hash(TEST_USER_PASSWORD),
                )
                user = crud_user.create(db=db, user=user_in)
    finally:
        db.close()
